#define _POSIX_C_SOURCE 199309L

#define RLITE_CRLF "\r\n"
#define RLITE_STR_OK "OK"
#define RLITE_STR_ERR "ERR"
#define RLITE_PONG "PONG"
#define RLITE_QUEUED "QUEUED"
#define RLITE_WRONGTYPEERR "WRONGTYPE Operation against a key holding the wrong kind of value"
#define RLITE_WRONGNUMBEROFARGUMENTS "ERR wrong number of arguments for %s"
#define RLITE_NOKEYERR "ERR No such key"
#define RLITE_SYNTAXERR "ERR syntax error"
#define RLITE_SAMEOBJECTERR "ERR source and destination objects are the same"
#define RLITE_OUTOFRANGEERR "ERR index out of range"
#define RLITE_INVALIDMINMAXERR "ERR min or max not valid string range item"
#define RLITE_NOSCRIPTERR "NOSCRIPT No matching script. Please use EVAL."
