#ifndef _RL_STATUS_H
#define _RL_STATUS_H

#define RL_OK 0
#define RL_INVALID_STATE 1
#define RL_FOUND 2
#define RL_NOT_FOUND 3
#define RL_OUT_OF_MEMORY 4
#define RL_INVALID_PARAMETERS 5
#define RL_UNEXPECTED 6
#define RL_NOT_IMPLEMENTED 7
#define RL_WRONG_TYPE 8
#define RL_END 9
#define RL_DELETED 10
#define RL_NAN 11
#define RL_OVERFLOW 12
#define RL_OUTDATED 13

#endif
